"use strict";

requirejs('controls/control');
requirejs('controls/trackListView');
localRequirejs('three.min.js');
localRequirejs('3DController.js');

if (!window._FlowAlbumViewErrors)
    window._FlowAlbumViewErrors = []; // for error catching (to be removed later)

inheritClass('FlowAlbumView', Control, {
    initialize: function (elem, params) {
        params.dynamicSize = false;
        FlowAlbumView.$super.initialize.apply(this, arguments);
        
        const _this = this;
        this.isSearchable = true; // for search bar
        this.parent = elem;
        elem.classList.add('fill');
        elem.setAttribute('tabindex', '-1');

        // Settings
        this.loadSettings();

        // For default background color: Skins can override it by adding to their LESS:
        //  .threeDView-defaultStyles{ background-color: [color]!important; }
        let defaultStylesElem = document.createElement('div');
        defaultStylesElem.classList.add('threeDView-defaultStyles');
        elem.appendChild(defaultStylesElem);
        // This is a CSS computed style, so changes to the LESS/CSS will automatically be reflected in this object
        this.defaultStyles = getComputedStyle(defaultStylesElem);
        
        this._setCornerBoxText = (msg) => {
            if (this.settings.textBoxVisible) this.cornerTextBox.innerText = msg;
        }
        
        this.cornerTextBox = document.createElement('div');
        this.cornerTextBox.classList.add('threeDView-CornerTextBox');
        elem.appendChild(this.cornerTextBox);
        
        this.settingsButton = document.createElement('div');
        this.settingsButton.setAttribute('data-control-class', 'ToolButton');
        this.settingsButton.setAttribute('data-icon', 'options');
        this.settingsButton.setAttribute('data-tip', 'Options');
        this.settingsButton.classList.add('threeDView-settingsButton');
        initializeControl(this.settingsButton);
        this.settingsButton.controlClass.contextMenu = new Menu([{
            title: _('Preload Album Art (May take some time)'),
            execute: () => {
                if (_this.controller) _this.controller.preloadThumbnails();
            }
        }]);
        elem.appendChild(this.settingsButton);

        window.flow = this;

        ODS('FlowAlbumView - init');

        this.boundingRect = elem.getBoundingClientRect();

        this.camera = new THREE.PerspectiveCamera(50, this.boundingRect.width / this.boundingRect.height, 0.1, 1000);
        this.scene = new THREE.Scene();
        // Set the scene's background color
        this._setSceneBackground();

        ODS('FlowAlbumView - Creating controller')
        this.controller = new FlowController(this);

        // ThreeJS scene stuff
        this._setCameraPosition();
        const light = new THREE.PointLight(0xffffff, 2, 14, 1.5);
        light.position.set(0, 5, 3);
        this.scene.add(light);
        const ambientLight = new THREE.AmbientLight(0x484848); // soft white light
        this.scene.add(ambientLight);

        this.createNewContext();
        this.animationFrameTimeout = requestFrame(this.animate.bind(this));
        // this.triggerTemporaryForcedFrames(50);

        // Context menu
        //  The context menu population is handled by the rest of MM. 
        //  For tracklistmenus, it depends on the datasource's getSelected (set in 3DController)
        var _contextMenu = menus.createTracklistMenu(elem);
        this.contextMenu = () => {
            if (this.controller && this.controller._mousedOverObject && this.controller._mousedOverObject.mesh.visible) {
                return _contextMenu;
            }
        }
		
        this.onDoubleClick = function(e) {

            let mousedOverObject = _this.controller._mousedOverObject;
            // If object is in the center
            if (mousedOverObject && _this.controller.dataSource && Math.abs(mousedOverObject.position) < 0.1) {
                // Make sure the carousel is not in the middle of moving
                if (_this.controller.position - _this.controller.target < 0.2) {
                    let index = parseInt(-1 * _this.controller.target);
                    _this.controller.dataSource.locked(async () => {
                        let album = _this.controller.dataSource.getValue(index);
                        if (album) {
                            let tracklist = album.getTracklist();
                            await tracklist.whenLoaded();
                            if (_this.parentView.viewNode.handlerID === 'playlist') {
                                // only allow the songs that are IN the playlist to play (not the whole album)
                                tracklist = await _this.controller.filterFromPlaylist(tracklist);
                            }
                            await player.stopAsync();
                            await player.addTracksAsync(tracklist, { withClear: true });
                            player.playAsync();
                        }
                        else ODS('no album');
                    });
                }
            }
        }
        
        // This is to open the settings menu with the button
        var addonList = app.getAddonList('views');
        addonList.whenLoaded().then(() => {
            addonList.forEach(itm => {
                if (itm.ext_id.toLowerCase() === '3d album view') {
                    this.addon = itm;
                }
            })
        })
        
        this._handleSettingsChange = function() {
            this.loadSettings();
        }.bind(this);

        this.localListen(elem, 'keydown', handleKeypress, true);
        this.localListen(elem, 'mousemove', onMouseMove, false);
        this.localListen(elem, 'layoutchange', onLayoutChange);
        this.localListen(this.container, 'datasourcechanged', onDataSourceChanged);
        this.localListen(this.renderer.domElement, 'webglcontextlost', onContextLost);
        this.localListen(elem, 'dblclick', this.onDoubleClick);
        this.localListen(elem, 'wheel', onMouseWheel);
        app.listen(app, 'settingschange', this._handleSettingsChange);
        app.listen(app, 'playlistchange', (playlist) => {
            if (this.playlist && playlist.id === this.playlist.id) {
                clearTimeout(this.playlistChangeTimeout);
                this.playlistChangeTimeout = requestTimeout(async () => {
                    ODS('FlowAlbumView - Playlist changed; updating DS');
                    let tracklist = await _this.playlist.getTracklist();
                    if (tracklist.getAlbumList) {
                        let newDS = await tracklist.getAlbumList();
                        // sanity check to make sure the view's playlist hasn't changed
                        if (_this.playlist && playlist.id === _this.playlist.id) {
                            _this.dataSource = newDS;
                            // _this.completeTracklist = tracklist; This line also has to be fired under datasourcechanged, in case user switches straight from a different view to a playlist
                        }
                    }
                    else throw new Error('getAlbumList undefined!!');
                }, 2000); // wait 2s for another change, if there isn't any then update the data source
            }
        })
        this.localListen(this.settingsButton, 'click', () => {
            if (_this.addon) {
                uitools.openDialog('dlgAddonConfig', {
                    modal: true,
                    addon: _this.addon
                }, () => {
                    
                });
            }
        });
        this.localListen(window, 'lesschange', () => {
            this.defaultStyles = getComputedStyle(defaultStylesElem);
            this.loadSettings();
        });
        
        _this.resizeTicking = false;

        function onMouseWheel(e) {
            _this.controller.onMouseWheel(e);
        }

        function onMouseMove(e) {
            // calculate mouse position in normalized device coordinates
            // (-1 to +1) for both components

            _this.controller.mouse.x = ((e.clientX - _this.boundingRect.left) / _this.boundingRect.width) * 2 - 1;
            _this.controller.mouse.y = - ((e.clientY - _this.boundingRect.top) / _this.boundingRect.height) * 2 + 1;
        }

        function handleKeypress(e) {
            var handled = false;

            if (e.key === 'ArrowLeft') {
                _this.controller.moveRight();
                handled = true;
            }
            else if (e.key === 'ArrowRight') {
                _this.controller.moveLeft();
                handled = true;
            }
        }

        function onLayoutChange() {
            // console.log('Layout change');
            _this.onResize();
            _this.triggerTemporaryForcedFrames(20);
        }

        async function onDataSourceChanged() {
            if (_this.controller) _this.controller.setDataSource(_this._dataSource);
            
            if (this.playlist) {
                let tracklist = await _this.playlist.getTracklist();
                _this.completeTracklist = tracklist; // #18722
            }
        }

        // Handler for WebGL context loss event
        function onContextLost() {
            console.log('contextLost');
            if (!document.hidden) {
                console.log('document not hidden');
                restoreContext();
            }
            else {
                app.listen(document, 'visibilitychange', restoreContext);
            }
        }

        function restoreContext() {
            // For some reason, the WebGL context loses its WEBGL_lose_context extension, 
            //  which ThreeJS depends on to do a forceContextRestore() event. Therefore, it is best to destroy and re-create
            //  the renderer.

            if (document.visibilityState === 'visible') {
                _this.createNewContext();
                
                cancelAnimationFrame(_this.animationFrameTimeout);
                _this.animationFrameTimeout = requestFrame(_this.animate.bind(_this));

                app.unlisten(document, 'visibilitychange', restoreContext);
            }
            else {
                console.log('Document not visible; doing nothing');
            }
        }
    },
    _incrementalSearchMessageSuffix: function () {
        return ''; // not supported
    },
    // for "scrolls to match" searching
    performIncrementalSearch(searchPhrase) {
        if (!searchPhrase || searchPhrase == '') return;

        this.controller.scrollToLetter(searchPhrase);

        return true;
    },
    // for search bar filtering
    filterSource(searchPhrase) {
        // If the search phrase is empty, un-filter the albums
        if (!searchPhrase || searchPhrase == '' || searchPhrase == ' ') {
            this.controller.resetFilter();
            return;
        }
        this.controller.filterAlbums(searchPhrase);
    },
    // Required for the search bar
    showToast: function (message) {
        uitools.toastMessage.show(message, {
            disableUndo: true,
            disableClose: true,
            delay: 3000,
        });
    },
    loadSettings() {
        // To show a tooltip if the thumbnail setting has changed
        if (this.settings) 
            var prevThumbnailSetting = this.settings.highResThumbnails; 

        this.settings = app.getValue('FlowAlbumView_settings', {
            spacingMultiplier: 1,
            disablePerspectiveDistortion: false,
            capFramerate: false,
            textBoxVisible: true,
            highResThumbnails: true,
            backgroundColor: undefined,
            useHotlinks: true,
            throttleCarouselSpeed: true,
            mouseWheelScrollNumber: undefined,
            textFields: {
                header: 'title',
                subheader: 'albumArtist',
                line1: 'date',
                line2: 'description'
            },
            textHotlinks: {
                header: 'album',
                subheader: undefined,
                line1: 'year',
                line2: undefined
            }
        });
        if (this.controller) this.controller.handleSettingsChange();
        // Update the scene's background color
        if (this.scene) {
            this._setSceneBackground();
        }
        if (this.camera) {
            this._setCameraPosition();
        }
        if ((typeof prevThumbnailSetting !== undefined) && prevThumbnailSetting != this.settings.highResThumbnails) {
            if (this.controller && this.controller.albumArts) this.controller.albumArts.clear(true)
        }
    },
    _setSceneBackground() {
        let rgb = this.settings.backgroundColor;
        if (rgb) this.scene.background = new THREE.Color(`rgb(${rgb[0]},${rgb[1]},${rgb[2]})`);
        else this.scene.background = new THREE.Color(this.defaultStyles.backgroundColor);
    },
    cleanUp() {
        console.log('cleanUp called');

        // Unlisten events and cancel the render frame
        app.unlisten(this.parent);
        app.unlisten(this.container);
        cancelAnimationFrame(this.animationFrameTimeout);

        // Clean up and dispose of THREE.js objects.
        this.controller.cleanUp();
        this.renderer.dispose();

        this.renderer = null;
        this.controller = null;
        this.scene = null;
        this.camera = null;
        // window.flow = null;

        FlowAlbumView.$super.cleanUp.call(this);
    },
    createNewContext() {
        const _this = this;
        try {
            cancelAnimationFrame(this.animationFrameTimeout);

            if (this.renderer) {
                let oldCanvas = this.renderer.domElement;
                this.parent.removeChild(oldCanvas);
                app.unlisten(this.renderer.domElement, 'mousedown', onMouseDown);
            }
            console.log('Creating new context');

            // this.renderer = null;
            // console.log('creating tempRenderer');
            var tempRenderer = new THREE.WebGLRenderer({ antialias: true });
            // console.log('created tempRenderer');
            this.renderer = tempRenderer;

            this.renderer.setSize(this.boundingRect.width, this.boundingRect.height);
            this.parent.appendChild(this.renderer.domElement);

            // this.controls = new OrbitControls( this.camera, this.renderer.domElement );
            
            this.localListen(this.renderer.domElement, 'mousedown', onMouseDown);
        }
        catch (err) {
            this._setCornerBoxText(err.toString());
            console.error(err);
        }

        function onMouseDown(e) {
            window.lastFocusedControl = _this.parent;
            _this.controller.onMouseDown(e);
        }
    },
    _setCameraPosition() {
        var ratio = this.camera.aspect;
        
        if (this.settings.disablePerspectiveDistortion) {
            this.camera.position.y = 2;
            this.camera.position.z = 9.5;
        }
        else {
            this.camera.position.y = 3.8;
            this.camera.position.z = 9.5;
        }
        
        // Update the camera's fov and angle, and album spacing, depending on the view's aspect ratio
        if (ratio > 2.5) {
            let ratioDiff = ratio - 2.5;
            this.controller.SPACING = Math.min(1.5 + ratioDiff / 2.2, 3.1) * this.settings.spacingMultiplier;
            this.camera.fov = Math.max(50 - ratioDiff * 4.5, 30);
            this.camera.lookAt(0, Math.min(-0.5 + ratioDiff / 5, 0.075), 0);
        }
        else {
            this.controller.SPACING = 1.5 * this.settings.spacingMultiplier;
            this.camera.fov = 50;
            this.camera.lookAt(0, -0.5, 0);
        }
    },
    onResize() {
        if (!this.resizeTicking) {
            this.requestTimeout(() => {
                if (this.controller.enabled) {
                    this.boundingRect = this.parent.getBoundingClientRect();

                    const aspect = this.boundingRect.width / this.boundingRect.height;
                    this.camera.aspect = aspect;
                    this.camera.updateProjectionMatrix();
                    this._setCameraPosition();
                    
                    this.renderer.setSize(this.boundingRect.width, this.boundingRect.height);
                    
                    this.resizeTicking = false;

                    this.controller.updateDOMControlSize();
                    this.controller.updateTextSize(this.boundingRect.height);
                }
            }, 16, 'resize');
            this.resizeTicking = true;
        }
    },
    animate() {
        if (this.renderer && this.controller) {

            let didRender = this.controller.animate();

            let st = Date.now();
            let end = st;
            if (didRender) {
                this.renderer.render(this.scene, this.camera);
                end = Date.now();
                
                let prevLastFrameTime = this.prevLastFrameTime;
                let lastFrameTime = this.lastFrameTime;
                this.lastFrameTime = end;
                this.prevLastFrameTime = lastFrameTime;
                
                // this._setCornerBoxText(`${parseInt(2000 / ((end - lastFrameTime) + (lastFrameTime - prevLastFrameTime)))} fps`);
                // stats.update();
            }
            
            // Force a smooth resize if necessary
            if (didRender && this.forceSmoothResize) {
                this.onResize();
            }
            
            // if (end - st > 25) {
            //     ODS('flow - Long render time')
            //     console.log(`Warning: Long render time; time: ${(end - st).toFixed(3)} ms`);
            // }

            if (this.controls) this.controls.update();
            
            cancelAnimationFrame(this.animationFrameTimeout);
            this.animationFrameTimeout = requestFrame(this.animate.bind(this));
        }
    },
    /**
     * Temporarily force the view to render every frame for a short amount of time, to provide a smoother experience during important transitions.
     * @param {number} [duration] Duration in ms
     */
    triggerTemporaryForcedFrames(duration) {
        if (!duration) duration = 100;
        if (this.controller) {
            this.controller.forceRefresh = true;
            requestTimeout(() => {if (this.controller) this.controller.forceRefresh = false}, duration);
        }
    },
    /**
     * 
     * @param {number} [duration] Duration in ms
     */
    triggerSmoothResize(duration) {
        if (!duration) duration = 100;
        if (this.controller) {
            this.forceSmoothResize = true;
            requestTimeout(() => {this.forceSmoothResize = false}, duration);
        }
        this.triggerTemporaryForcedFrames(duration);
    },
    // I don't want to load the tracklist inside the controller when the tracklist isn't loaded,
    //  so this is a way to force the controller to load the tracklist when the subview is shown
    refreshTracklistDS() {
        this.controller._updateCurrentAlbum(true);
    },
    // Clear the tracklist data source from mem
    clearTracklistDS() {
        this._tracklistDataSource = app.utils.createTracklist();
    }
}, {
    /**
     Gets/sets the datasource which is/will be shown
        NOTE: This data source is never filtered. The filtering is done inside 3DController. To access filtered DS, use this.controller.dataSource.

    @property dataSource
    @type object
    */
    dataSource: {
        get: function () {
            return this.controller ? this.controller.dataSource : this._dataSource;
        },
        set: function (ds) {
            if (this._dataSource == ds)
                return;

            var oldDataSource = this._dataSource;
            if (this._dataSource) {
                // do stuff for old data source
            }
            this._dataSource = ds;
            if (this._dataSource) {
                // do stuff for new data source
                this._dataSource.getTracklist = function () {
                    return app.utils.createTracklist();
                }
            }
            
            if (this.parentView.viewNode.handlerID === 'playlist') {
                this.isPlaylist = true;
            }
            else this.isPlaylist = false;

            var evt;
            evt = createNewCustomEvent('datasourcechanged', {
                detail: {
                    newDataSource: this._dataSource,
                    oldDataSource: oldDataSource
                },
                bubbles: true,
                cancelable: true
            });
            this.container.dispatchEvent(evt);
        }
    },
    /**
     * This is for when it's a playlist, so we can filter the tracklist within the playlist
     */
    completeTracklist: {
        get: function () {
            return this._completeTracklist;
        },
        set: function(newList) {
            this._completeTracklist = newList;
            // tracklistIDs will resolve when, as the name implies, the tracklist IDs are ready
            this.tracklistIDs = new Promise((resolve, reject) => {
                this._completeTracklist.whenLoaded()
                .then(() => {
                    let tracklistIDsArr;
                    
                    // post 5.0.2
                    if (typeof this._completeTracklist.getAllValues === 'function') {
                        console.log('Post 5.0.2');
                        tracklistIDsArr = new Int32Array(this._completeTracklist.getAllValues('idsong'));
                        resolve(tracklistIDsArr);
                    }
                    // pre 5.0.2
                    else {
                        tracklistIDsArr = new Int32Array(this._completeTracklist.count);
                        console.log('Pre 5.0.2');
                        this._completeTracklist.locked(() => {
                            let song;
                            for (let i = 0; i < tracklistIDsArr.length; ++i) {
                                song = this._completeTracklist.getFastObject(i, song);
                                tracklistIDsArr[i] = song.idsong;
                            }
                            resolve(tracklistIDsArr);
                        });
                    }
                });
            });
        }
    },
    // whether the controller needs to load the tracklist
    needsTracklist: {
        get: function () {
            return (!!this.parentView.tracklistControl);
        },
        set: function () {
            
        }
    },
    tracklistDataSource: {
        get: function () {
            return this._tracklistDataSource;
        },
        set: function (ds) {
            this._tracklistDataSource = ds;
            // Assign the data source to the tracklist control
            if (this.parentView.tracklistControl) {
                this.parentView.tracklistControl.dataSource = ds;
            }
        }
    }
});