/**
@module UI snippets
*/

"use strict";
requirejs('controls/control');

/**
UI color picker control.

@class ColorPicker
@constructor
@extends Control
*/
inheritClass('TempColorPicker', Control, {
    initialize: function (elem, params) {
        TempColorPicker.$super.initialize.apply(this, arguments);
        
        if (!params) params = {};
        
        const _this = this;
        
        // Default size: 300px. Can be controlled with params.size.
        var WIDTH;
        if (params.size && !isNaN(params.size)) WIDTH = parseInt(params.size);
        else WIDTH = 300;
        
        var hueWidth = Math.round(WIDTH * 0.2);
        var prvWidth = Math.round(WIDTH * 0.5)
        var prvHeight = Math.round(WIDTH * 0.3);
        var pad = 7;
        var sqPad = 5;

        this.svCanv = document.createElement('canvas');
        this.hueCanv = document.createElement('canvas');
        this.prvCanv = document.createElement('canvas');
        this.edit = document.createElement('div');
        var editContainer = document.createElement('div');
        var rest = document.createElement('div');
        var container = document.createElement('div');
        
        // Sat-Val (square)
        this.svCanv.width = WIDTH;
        this.svCanv.height = WIDTH;
        this.svCanv.classList.add('floatLeft');
        // Hue (rectabgle)
        this.hueCanv.height = WIDTH;
        this.hueCanv.width = hueWidth;
        this.hueCanv.classList.add('floatLeft');
        // Preview
        this.prvCanv.width = 1;
        this.prvCanv.height = 1;
        this.prvCanv.style.width = prvWidth;
        this.prvCanv.style.height = prvHeight;
        this.prvCanv.style.padding = pad;
        // Text edit
        editContainer.style.padding = pad;
        editContainer.style.width = prvWidth;
        this.edit.setAttribute('data-control-class', 'Edit');
        container.style.position = 'relative';
        container.classList.add('floatRow');
        rest.classList.add('floatLeft');
        elem.classList.add('colorPicker');

        container.appendChild(this.svCanv);
        container.appendChild(this.hueCanv);
        container.appendChild(rest);
        rest.appendChild(this.prvCanv);
        rest.appendChild(editContainer);
        editContainer.appendChild(this.edit);
        elem.appendChild(container);
        
        initializeControl(this.edit);

        var svCtx = this.svCanv.getContext('2d');
        var hueCtx = this.hueCanv.getContext('2d');
        var prvCtx = this.prvCanv.getContext('2d');

        var grdRed, grdGreen, grdBlue, grdWhite, grdBlack;
        
        var animationFrameTimeout;
        var boundingRectUpdateNeeded;
        
        var hexValue;
        var hsv = {h: 0.5, s: 0.5, v: 0.5};
        // Default color: Middle blueish. Can be customized with params.value (hex value), params.rgb (array, 0-255), or params.hsv (array, 0-1).
        if (params.value && typeof params.value === 'string') {
            let rgb = hexToRgb(params.value);
            if (rgb) hsv = rgbToHsv(rgb.r, rgb.g, rgb.b);
        }
        if (params.rgb && params.rgb[0] && params.rgb[1] && params.rgb[2]) {
            hsv = rgbToHsv(params.rgb[0], params.rgb[1], params.rgb[2]);
        }
        if (params.hsv && params.hsv[0] && params.hsv[1] && params.hsv[2]) {
            hsv.h = Math.min(Math.max(params.hsv[0], 0), 1);
            hsv.s = Math.min(Math.max(params.hsv[1], 0), 1);
            hsv.v = Math.min(Math.max(params.hsv[2], 0), 1);
        }
        
        // When the document scrolls, the bounding rect will need to be updated.
        this._handleScroll = function(e) {
            boundingRectUpdateNeeded = true;
        }.bind(this);
        
        this._svSeekHandler = function(e) {
            let rect = _this.svRect;
            let x = e.clientX, y = e.clientY;
            
            let svX, svY;
            
            // too far to the top
            if (y < rect.top + pad)
                svY = 0;
            // too far to the bottom
            else if (y > rect.top + rect.height - pad)
                svY = 1;
            else 
                svY = (y - rect.top - pad) / (WIDTH-pad*2);
            
            // too far to the left
            if (x < rect.left + pad)
                svX = 0;
            // too far to the bottom
            else if (x > rect.x + rect.width - pad)
                svX = 1;
            else
                svX = (x - rect.left - pad) / (WIDTH-pad*2);
            
            hsv.s = svX;
            hsv.v = 1 - svY;
            
            // request a new frame
            if (!animationFrameTimeout) {
                animationFrameTimeout = requestFrame(() => {
                    render();
                    _this._updateEditValue();
                });
            }
        };
        
        this._hueSeekHandler = function(e) {
            let rect = _this.hueRect;
            let y = e.clientY;
            
            let huePos;
            
            // too far to the top
            if (y < rect.top)
                huePos = 0;
            // too far to the bottom
            else if (y > rect.top + rect.height)
                huePos = 1;
            else
                huePos = (y - rect.top) / WIDTH;
            
            hsv.h = huePos;
            
            // request a new frame
            if (!animationFrameTimeout) {
                animationFrameTimeout = requestFrame(() => {
                    render();
                    _this._updateEditValue();
                });
            }
        };
        
        this._updateBoundingRect = function() {
            _this.hueRect = _this.hueCanv.getBoundingClientRect();
            _this.svRect = _this.svCanv.getBoundingClientRect();
        };
        
        function svSeekStart(e) {
            if (elem.hasAttribute('data-disabled') && !!elem.getAttribute('data-disabled')) return;
            app.listen(document, 'mousemove', _this._svSeekHandler);
            // Update the bounding rect if needed first
            if (boundingRectUpdateNeeded) {
                _this._updateBoundingRect();
                boundingRectUpdateNeeded = false;
            }
            _this._svSeekHandler(e);
        }
        
        function svSeekEnd(e) {
            app.unlisten(document, 'mousemove', _this._svSeekHandler);
        }
        
        
        function hueSeekStart(e) {
            if (elem.hasAttribute('data-disabled') && !!elem.getAttribute('data-disabled')) return;
            app.listen(document, 'mousemove', _this._hueSeekHandler);
            // Update the bounding rect if needed first
            if (boundingRectUpdateNeeded) {
                _this._updateBoundingRect();
                boundingRectUpdateNeeded = false;
            }
            _this._hueSeekHandler(e);
        }
        
        function hueSeekEnd() {
            app.unlisten(document, 'mousemove', _this._hueSeekHandler);
        }
        
        // Perform the necessary one-time setup
        function init() {
            // hue line
            grdRed = hueCtx.createLinearGradient(pad, pad, pad, WIDTH-pad);
            grdRed.addColorStop(0.16666, '#FF0000FF');
            grdRed.addColorStop(0.33333, '#FF000000');
            grdRed.addColorStop(0.66666, '#FF000000');
            grdRed.addColorStop(0.83333, '#FF0000FF');
            grdGreen = hueCtx.createLinearGradient(pad, pad, pad, WIDTH-pad);
            grdGreen.addColorStop(0, '#00FF0000');
            grdGreen.addColorStop(0.16666, '#00FF00FF');
            grdGreen.addColorStop(0.5, '#00FF00FF');
            grdGreen.addColorStop(0.66666, '#00FF0000');
            grdBlue = hueCtx.createLinearGradient(pad, pad, pad, WIDTH-pad);
            grdBlue.addColorStop(0.33333, '#0000FF00');
            grdBlue.addColorStop(0.5, '#0000FFFF');
            grdBlue.addColorStop(0.83333, '#0000FFFF');
            grdBlue.addColorStop(1, '#0000FF00');
            // B/W sat/val square
            grdWhite = svCtx.createLinearGradient(pad, pad, WIDTH-pad, pad);
            grdWhite.addColorStop(0, '#FFFFFFFF');
            grdWhite.addColorStop(1, '#FFFFFF00');
            grdBlack = svCtx.createLinearGradient(pad, pad, pad, WIDTH-pad);
            grdBlack.addColorStop(0, '#00000000');
            grdBlack.addColorStop(1, '#000000FF');

            render();
            _this._updateBoundingRect();
        }
        
        function render() {
            hue();
            square();
            preview();
            drawHueRect();
            drawSVRect();
            function hue() {
                hueCtx.clearRect(0, 0, hueWidth, WIDTH);
                hueCtx.globalCompositeOperation = 'lighter';
                hueCtx.fillStyle = grdRed;
                hueCtx.fillRect(pad, pad, hueWidth-pad*2, WIDTH-pad*2);
                hueCtx.fillStyle = grdGreen;
                hueCtx.fillRect(pad, pad, hueWidth-pad*2, WIDTH-pad*2);
                hueCtx.fillStyle = grdBlue;
                hueCtx.fillRect(pad, pad, hueWidth-pad*2, WIDTH-pad*2);
            }
            function square() {
                svCtx.clearRect(0, 0, WIDTH, WIDTH);
                
                let thisRGB = hsvToRgb(hsv.h, 1, 1);
                svCtx.fillStyle = rgb(thisRGB.r, thisRGB.g, thisRGB.b);
                svCtx.fillRect(pad, pad, WIDTH-pad*2, WIDTH-pad*2);
                
                svCtx.fillStyle = grdWhite;
                svCtx.fillRect(pad, pad, WIDTH-pad*2, WIDTH-pad*2);
                svCtx.fillStyle = grdBlack;
                svCtx.fillRect(pad, pad, WIDTH-pad*2, WIDTH-pad*2);
            }
            function preview() {
                let thisRGB = hsvToRgb(hsv.h, hsv.s, hsv.v);
                prvCtx.fillStyle = rgb(thisRGB.r, thisRGB.g, thisRGB.b);
                prvCtx.fillRect(0, 0, 1, 1);
            }
            function rgb(r, g, b) {
                return 'rgb(' + String(r) + ',' + String(g) + ',' + String(b) + ')';
            }
            function drawHueRect() {
                hueCtx.globalCompositeOperation = 'source-over';
                var pixelPos = lerp(pad, WIDTH-pad, hsv.h);
                // black surrounding stroke
                hueCtx.strokeStyle = 'black';
                hueCtx.lineWidth = 4;
                hueCtx.strokeRect(pad, pixelPos-sqPad, hueWidth-pad*2, sqPad*2);
                // white box
                hueCtx.strokeStyle = 'white';
                hueCtx.lineWidth = 2;
                hueCtx.strokeRect(pad, pixelPos-sqPad, hueWidth-pad*2, sqPad*2);
            }
            
            function drawSVRect() {
                svCtx.globalCompositeOperation = 'source-over';
                var pixelX = lerp(pad, WIDTH-pad, hsv.s);
                var pixelY = lerp(pad, WIDTH-pad, (1 - hsv.v));
                // black surrounding stroke
                svCtx.strokeStyle = 'black';
                svCtx.lineWidth = 4;
                svCtx.strokeRect(pixelX-sqPad, pixelY-sqPad, sqPad*2, sqPad*2);
                // colored inside
                let thisRGB = hsvToRgb(hsv.h, hsv.s, hsv.v);
                svCtx.fillStyle = rgb(thisRGB.r, thisRGB.g, thisRGB.b);
                svCtx.fillRect(pixelX-sqPad, pixelY-sqPad, sqPad*2, sqPad*2);
                // white box
                svCtx.strokeStyle = 'white';
                svCtx.lineWidth = 2;
                svCtx.strokeRect(pixelX-sqPad, pixelY-sqPad, sqPad*2, sqPad*2);
            }
            
            animationFrameTimeout = undefined;
        }
        
        /**
         * Update the value of the color picker
         * @param {string} value 
         */
        this._setValue = function(value) {
            var rgb = hexToRgb(value);
            // Only update and render if we have an interpretable value
            if (rgb) {
                hsv = rgbToHsv(rgb.r, rgb.g, rgb.b);
                if (value != hexValue) render();
                
                // If the contents don't include a # at the start, add it
                if (!value.startsWith('#')) {
                    _this.edit.controlClass.value = '#' + value;
                }
            }
        }
        
        /**
         * Internal function to update the edit's value when the HSV variable changes.
         */
        this._updateEditValue = function(alsoRender) {
            var rgb = hsvToRgb(hsv.h, hsv.s, hsv.v);
            var hex = rgbToHex(rgb.r, rgb.g, rgb.b);
            hexValue = hex;
            _this.edit.controlClass.value = hex;
        }
        
        this._getValue = function() {
            var rgb = hsvToRgb(hsv.h, hsv.s, hsv.v);
            return rgbToHex(rgb.r, rgb.g, rgb.b);
        }
        
        this._getHSV = function () {
            return [hsv.h, hsv.s, hsv.v];
        }
        
        this._setHSV = function (h, s, v) {
            hsv.h = Math.min(Math.max(h, 0), 1);
            hsv.s = Math.min(Math.max(s, 0), 1);
            hsv.v = Math.min(Math.max(v, 0), 1);
            _this._updateEditValue();
            render();
        }
        
        this._getRGB = function () {
            let rgb = hsvToRgb(hsv.h, hsv.s, hsv.v);
            return [rgb.r, rgb.g, rgb.b];
        }
        
        this._setRGB = function (r, g, b) {
            hsv = rgbToHsv(r, g, b);
            _this._updateEditValue();
            render();
        }
        
        /**
         * @param {number} r 
         * @param {number} g 
         * @param {number} b 
         * @returns {string}
         */
        function rgbToHex(r, g, b) {
            // Clamp r, g, and b
            r = Math.min(Math.max(r, 0), 255);
            g = Math.min(Math.max(g, 0), 255);
            b = Math.min(Math.max(b, 0), 255);
            // Turn them into hexadecimal
            r = Math.ceil(r).toString(16).padStart(2,'0');
            g = Math.ceil(g).toString(16).padStart(2,'0');
            b = Math.ceil(b).toString(16).padStart(2,'0');
            return `#${r}${g}${b}`;
        }
        
        /**
         * @param {string} hex
         * @returns {object}
         */
        function hexToRgb(hex) {
            // Remove the # if needed
            if (hex.startsWith('#')) hex = hex.substring(1, hex.length);
            var r, g, b;
            // 3-length hex codes just have an 8-bit color instead of 16-bit
            if (hex.length == 3) {
                r = parseInt(hex[0], 16) * 17;
                g = parseInt(hex[1], 16) * 17;
                b = parseInt(hex[2], 16) * 17;
            }
            // 6-length hex codes are the standard
            else if (hex.length >= 6) {
                r = parseInt(hex.substring(0, 2), 16);
                g = parseInt(hex.substring(2, 4), 16);
                b = parseInt(hex.substring(4, 6), 16);
            }
            else {
                // if the hex's length is 1, 2, or 4, we have no way to interpret it
                return;
            }
            return {r, g, b};
        }

        /**
         * @param {number} r 
         * @param {number} g 
         * @param {number} b 
         * @returns {object}
         */
        function rgbToHsv(r, g, b) {
            // Credit to MERAMU KODING https://en.meramukoding.com/js-converting-rgb-colorspace-to-hsv/
            var h, s, v;

            var maxColor = Math.max(r, g, b);
            var minColor = Math.min(r, g, b);
            var d = maxColor - minColor;

            if (d === 0) h = 0;
            else if (r === maxColor) h = (6 + (g - b) / d) % 6;
            else if (g === maxColor) h = 2 + (b - r) / d;
            else if (b === maxColor) h = 4 + (r - g) / d;
            else h = 0;
            h = h / 6;

            if (maxColor !== 0) s = d / maxColor;
            else s = 0;

            v = maxColor / 255;

            return { h: h, s: s, v: v };
        };
        
        function hsvToRgb(h, s, v) {
            // Credit to EasyRGB: http://www.easyrgb.com/en/math.php
            //H, S and V input range = 0 ÷ 1.0
            //R, G and B output range = 0 ÷ 255
            var r, g, b, var_h, var_i, var_1, var_2, var_3, var_r, var_g, var_b;

            if (s == 0) {
                r = v * 255
                g = v * 255
                b = v * 255
            }
            else {
                var_h = h * 6
                if (var_h == 6) var_h = 0;
                var_i = parseInt(var_h);
                var_1 = v * (1 - s);
                var_2 = v * (1 - s * (var_h - var_i));
                var_3 = v * (1 - s * (1 - (var_h - var_i)));

                if (var_i == 0) { var_r = v; var_g = var_3; var_b = var_1 }
                else if (var_i == 1) { var_r = var_2; var_g = v; var_b = var_1 }
                else if (var_i == 2) { var_r = var_1; var_g = v; var_b = var_3 }
                else if (var_i == 3) { var_r = var_1; var_g = var_2; var_b = v }
                else if (var_i == 4) { var_r = var_3; var_g = var_1; var_b = v }
                else { var_r = v; var_g = var_1; var_b = var_2 }

                r = Math.round(var_r * 255);
                g = Math.round(var_g * 255);
                b = Math.round(var_b * 255);
            }
            return { r: r, g: g, b: b };
        }
        function lerp(a, b, t) {
            return (1-t)*a + t*b;
        }
        
        this.localListen(this.hueCanv, 'mousedown', (e) => hueSeekStart(e));
        this.localListen(this.svCanv, 'mousedown', (e) => svSeekStart(e));
        this.localListen(document.body, 'mouseup', () => {
            hueSeekEnd();
            svSeekEnd();
        });
        this.localListen(window, 'blur', () => {
            hueSeekEnd();
            svSeekEnd();
        });
        this.localListen(this.edit.controlClass, 'change', () => {
            var editValue = this.edit.controlClass.value;
            if (editValue) this._setValue(editValue);
        });
        app.listen(this.container, 'layoutchange', this._updateBoundingRect);
        app.listen(window, 'scroll', this._handleScroll, true);
        
        // 
        init();
        render();
        this._updateEditValue();
    },
    cleanUp() {
        app.unlisten(this.container, 'layoutchange', this._updateBoundingRect);
        app.unlisten(window, 'scroll', this._handleScroll, true);
        
        TempColorPicker.$super.cleanUp.call(this);
    },
}, {
    /**
        Get/set the hexadecimal value of the color picker. Can be a 3-digit hex value (#FFF) or a 6-digit hex value (#FFFFFF). The # is not necessary when setting; it is added automatically.
        @property value
        @type string
    */
    value: {
        get: function () {
            return this._getValue();
        },
        set: function (value) {
            this._setValue(value);
        }
    },
    /**
     * Get/set the HSV value of the color picker. HSV ranges from 0 to 1 on all values. Set with an array [h, s, v].
     */
    hsv: {
        get: function () {
            return this._getHSV();
        },
        set: function(hsv) {
            this._setHSV(hsv[0], hsv[1], hsv[2]);
        }
    },
    /**
     * Get/set the RGB value of the color picker. RGB ranges from 0 to 255 on all values. Set with an array [r, g, b].
     */
    rgb: {
        get: function () {
            return this._getRGB();
        },
        set: function(rgb) {
            this._setRGB(rgb[0], rgb[1], rgb[2]);
        }
    }
});